#!python

"""
Manipulate the configuration of an SMA appliance through it's Functional API,
using a library of JSON template files. Each is valid somewhere in the API schema,
and "somewhere" is given in the name of the file per below.

Each template name is broken into fields by a double underscore. Only
the first field is required; if a field is missing the
filename may contain zero underscores (only the first field is
present), a trailing double underscore (the field preceding the missing field
is present) or four underscores in
succession (the field preceding the missing field is present, and
possibly the one succeeding it). Note that only two fields are defined ATM.

The first field gives the API path location, substituting commas for
slashes. Eg: Setup,Eula or Console,SSH

If two fields are present, the second is the HTTP method to use with the
template (default if missing or empty is PUT).
"""

import argparse, requests, os, sys, re, string, json, textwrap, urllib3, time
import jsonpath_rw as jsonpath
from collections import OrderedDict
from pkg_resources import resource_listdir, resource_string, resource_filename
from getpass import getpass

requests.packages.urllib3.disable_warnings()

class Failure(Exception):
    pass
class APIFailure(Failure):
    pass
class ApplyFailure(Failure):
    pass
class MissingTemplate(Failure):
    pass
class TemplateContentError(Failure):
    pass
class CommandListError(Exception):
    pass

def find_support_dir(core):
    """Find the ~/.appli support dir, where runtime support files are kept.
    This may also be the virtualenv dir, if there is one.
    """
    support = os.path.expanduser(os.path.join("~", '.' + core))
    if os.path.isdir(support):
        return support
    return None

def decode_templatename(itemname):
    """Pick out and expand or supply defaults for the name fields, separated by
    double-underscores: location (in API, with commas replacing
    slashes), method (PUT by default).
    """
    try:
        (location0, method) = itemname.split('__', 1)
    except ValueError:
        location0 = itemname
        method = 'PUT'
         
    # Location assumes Console unless Console or Setup are specified.
    path = location0.split(',')
    if not (path[0] == 'Console' or path[0] == 'Setup'):
        path.insert(0, 'Console')
    location = '/'.join(path)

    return (method, location)

def get_template(templatename, userdir=None):
    """Find a template in user data, or failing that in package resources.
    Failing that raise IOError.
    """
    templatestring = None
    if userdir:
        filename = os.path.abspath(os.path.join(userdir, templatename))
        try:
            templatestring = open(filename).read()
        except IOError:
            pass
    if templatestring is None:
        # This will raise IOError if template doesn't exist:
        templatestring = resource_string('aventail.appli', 'appli-templates/%s' % templatename)
    return templatestring

def expand_template(templatestring, expansions):
    """Replace (( name:value )) strings in templatestring with matching
    [ (name, val), ] val from expansions, or with value. Names must be
    alphanumeric, one to eight chars long.
    """
    for e in expansions:
        n, v = e
        pattern = r'\( \( \s+ ' + n + r'\: [^)]*? \s+ \) \)'
        result = re.sub(pattern, v, templatestring, 0, re.X)
        templatestring = result

    # defaults
    result = re.sub(r'\( \( \s+ [a-zA-Z0-9_]{1,8} \: ([^) ]*?) \s+ \) \)', r'\g<1>',
                    templatestring, 0, re.X)
    return result

class Command(object):
    """
    Represent a command-line Command, like Version__GET, along with
    following template substitutions like name:value and HTTP query
    parameters like name=value.
    """
    def __init__(self, command_text):
        self.command = command_text
        self.tparams = []
        self.qparams = []
    def add_template_substitution(self, item):
        name, value = item.split(':', 1)
        self.tparams.append((name, value))
    def add_query_parameter(self, item):
        name, value = item.split('=', 1)
        self.qparams.append((name, value))    

class Applier:
    def __init__(self, args):
        """args: Hash-like object with values listed below"""

        self.subst_re = re.compile(r'\( \( ([^:]+) \: ([^)]*) \) \)', re.X)

        # Dir for template files. Default is also the virtualenv dir for appli.py,
        # and might contain the "settings" file for -p
        self.dir = args.dir
        if self.dir is None:
            raise Failure('No support dir found: ~/.appli')

        url_default = r'https://local\admin@192.168.0.10:8443/'
        persist_file = os.path.join(self.dir, "settings")
        persist = None
        try:
            if not args.ignore_persist:
                persist = json.loads(open(persist_file).read())
        except IOError:
            pass
        if persist is None:
            persist = {
                u'url': url_default,
            }
        
        if 'url' in args and args.url:
            url = args.url
        elif u'url' in persist:
            url = persist[u'url']
        else:
            url = url_default

        if 'hostname' in args:
            hostname = args.hostname
        else:
            hostname = None

        if args.prompt_password:
            if sys.stdin.isatty():
                typed_password = getpass(stream=sys.stderr)
            else:
                typed_password = sys.stdin.readline().rstrip()
        else:
            typed_password = None

        res = urllib3.util.parse_url(url)
        username = None
        password = None
        if res.auth:
            if ':' in res.auth:
                username, password = res.auth.split(':',1)
            else:
                username = res.auth
        self.scheme = res.scheme
        self.username = username
        self.password = typed_password or password
        self.hostname = hostname or res.hostname
        self.port = res.port or 8443
        self.path = res.path or '/'

        if args.persist:
            if self.username:
                if self.password:
                    auth = '%s:%s' % (self.username, self.password)
                else:
                    auth = self.username
            else:
                auth = None

            newu = urllib3.util.Url(
                scheme=self.scheme if self.scheme else 'https',
                auth=auth,
                host=self.hostname,
                port=self.port,
                path=self.path)

            try:
                os.unlink(persist_file)
            except OSError:
                pass
            try:
                os.mkdir(os.path.dirname(persist_file))
            except OSError:
                pass

            f = open(persist_file, "w")
            f.write(json.dumps( {u'url': newu.url} ))
            os.chmod(persist_file, 0600)
            f.close()
        
        self.timeout = float(args.timeout)	# 60 seconds
        self.failok = args.failok       # False
        self.no_user_templates = args.no_user_templates if 'no_user_templates' in args else False
        self.quiet = args.quiet         # False
        self.verbose = args.verbose     # False

        # Make .commands a list of template-name, tparams a parallel
        # list of name:value template-params, and qparams a parallel list of
        # name=value query-params. Note '=' vs ':'
        iscmd = lambda x: ':' not in x and '=' not in x and x[0] in string.ascii_uppercase
        self.commands = []
        if len(args.items):
            if not iscmd(args.items[0]):
                raise CommandListError("Start command list with a command!")
            for entry in args.items:
                if iscmd(entry):
                    self.commands.append(Command(entry))
                else:
                    if ':' in entry:
                        self.commands[-1].add_template_substitution(entry)
                    elif '=' in entry:
                        self.commands[-1].add_query_parameter(entry)
                    else:
                        raise CommandListError(entry)

    def apply(self):
        s = requests.Session()
        s.auth = (self.username, self.password)
        for cmd in self.commands:
            method, location = decode_templatename(cmd.command)
            userdir = self.dir if not self.no_user_templates else None
            try:
                templatestring = get_template(cmd.command, userdir=userdir)
            except IOError:
                if method != 'GET':
                    raise MissingTemplate(cmd.command)
                templatestring = ''
                
            if method == 'GET':
                body = ''
            else:
                body = expand_template(templatestring, cmd.tparams)

            url = "%s://%s:%u/%s" % (self.scheme, self.hostname, self.port, location)
            if cmd.qparams:
                url += '?' + '&'.join(map(lambda x: x[0] + '=' + x[1], cmd.qparams))
            if self.verbose:
                sys.stderr.write("%s %s\n" % (method, url))
            req = requests.Request(method, url, data=body)
            prepped = s.prepare_request(req)
            prepped.headers['Content-Type'] = 'application/json; charset=utf-8'
            try:
                resp = s.send(prepped, verify=False, timeout=(self.timeout, 60.0))
            except requests.exceptions.Timeout as e:
                if not self.quiet:
                    sys.stderr.write('timeout\n')
                resp = None
            except Exception:
                raise
            if not resp:
                if not self.failok:
                    raise APIFailure(location, method)
            if resp and resp.status_code > 299:
                if not self.quiet:
                    sys.stderr.write("%s\n", resp.headers)
                    sys.stderr.write("%s\n", resp.content)
                raise ApplyFailure(item, itemname, location, method)
            if resp:
                responsebody = resp.content  # else, by requests API docs, session leaks

                # Post-process GET requests - if template body isn't empty, make
                # (( jsonpath )) replacements and print the expansion.
                if method == 'GET':
                    sys.stdout.write(self.expand_response_body(responsebody, templatestring))

    def wait(self):
        redirections = 0
        timeouts = 0
        timeouts_after_redirection = 0
        start = int(time.time())
        try:
            method, location = decode_templatename(self.commands[0].command)
        except IndexError:
            location = 'Console/Version'
        url = "%s://%s:%u/%s" % (self.scheme, self.hostname, self.port, location)
        while int(time.time()) - start < int(self.timeout):
            try:
                resp = None
                resp = requests.head(
                    url,
                    allow_redirects=False,
                    verify=False,
                    timeout=(2.0, 5.0))
            except (requests.exceptions.Timeout, requests.exceptions.ConnectionError) as e:
                timeouts += 1
                if redirections > 0:
                    timeouts_after_redirection += 1
                sys.stdout.write('!')
                sys.stdout.flush()
                time.sleep(2)
                continue
            except Exception as e:
                sys.stdout.write('### Timed out on %s, check networking. Exception follows:\n' % url)
                raise
            if resp and resp.status_code >= 200 and resp.status_code <= 299:
                sys.stdout.write('(Ready.)\n')
                sys.exit(0)
            else:
                redirections += 1
                sys.stdout.write('.')
                sys.stdout.flush()
                time.sleep(2)

        if timeouts_after_redirection:
            sys.stdout.write('\n(Connected but not ready at first, then no connection. Did network address change?)\n')
        elif timeouts:
            sys.stdout.write('\n(%s timed out, check networking.)\n' % url)
        elif redirections:
            sys.stdout.write('\n(%s redirected, assuming Not Ready.)\n' % url)
        else:
            sys.stdout.write('\n(%s gave weird response; timeout too short?)\n' % url)
        sys.exit(1)

    def expand_response_body(self, responsebody, templatestring):
        """
        GET response bodies are assumed to be JSON. Empty GET templates
        specify writing the response body to stdout. Non-empty GET templates
        are themselves written to stdout, with (( jsonpath )) blocks replaced
        using values from the GET response object. Note open-paren, open-paren,
        space sequence, and space, close-paren, close-paren. (Space is any number
        of whitespace characters.)
        """
        open_re = re.compile(r'(.*?) \( \( \s+', re.X)
        close_re = re.compile(r'(.*?) \s+ \) \)', re.X)
        result = ''
        i = 0
        tlen = len(templatestring)
        if tlen == 0:
            return responsebody
        response = json.loads(responsebody)
        while i < tlen:
            mo_open = open_re.match(templatestring, pos=i)
            if not mo_open:
                result += templatestring[i:]
                return result
            s, i = mo_open.span()
            mo_close = close_re.match(templatestring, pos=i)
            if not mo_close:
                raise TemplateContentError("No close )) from pos=%d" % i)
            jp = jsonpath.parse(mo_close.group(1))
            result += templatestring[mo_open.start(1):mo_open.end(1)] + jp.find(response)[0].value
            i = mo_close.end()
        return result
                          
def list_default_templates():
    """Return a list of prepackaged templates."""
    return list(resource_listdir('aventail.appli', 'appli-templates'))

def do_list_templates(args):
    """Process the --list command-line option.
    Write a listing of templates to stdout.
    """
    res_filename = resource_filename('aventail.appli', 'appli-templates')
    if len(args.items):
        a = Applier(args)
        for cmd in a.commands:
            userdir = a.dir if not a.no_user_templates else None
            try:
                templatestring = get_template(cmd.command, userdir=userdir)
            except IOError:
                sys.stderr.write('No such template: %s\n' % cmd.command)
                sys.stderr.write('in module library location: %s\n' % res_filename)
                sys.stderr.write('or user templates location: %s\n' % args.dir)
                sys.exit(1)
            if not templatestring:
                sys.stdout.write('No template %s' % cmd.command)
            else:
                c = cmd.command
                if cmd.qparams:
                    c += '?' + '&'.join(map(lambda x: x[0] + '=' + x[1], cmd.qparams))
                sys.stdout.write('%s\n' % c)
                if cmd.tparams:
                    templatestring = expand_template(templatestring, cmd.tparams)
                sys.stdout.write(templatestring)
    else:
        sys.stdout.write('Standard distributed templates in %s:\n    ' % res_filename)
        sys.stdout.write('\n    '.join(list_default_templates()))
        if args.dir:
            contents = list(filter(lambda x: x[0] in string.ascii_uppercase,
                                   os.listdir(args.dir)))
            if len(contents):
                sys.stdout.write('\n')
                sys.stdout.write('User-supplied templates/overrides in %s:\n    ' % args.dir)
                sys.stdout.write('\n    '.join(contents))
        sys.stdout.write('\n')
        sys.exit(0)
    

def appli_func():
    """Command-line entry point for appli script.
    Parse command-line arguments, invoke operations they specify.
    """
    try:
        _appli_func(sys.argv)
    except MissingTemplate as e:
        sys.stderr.write("%s\n" % repr(e))
        sys.exit(1)
    except APIFailure as e:
        sys.stderr.write("AMC API failure. Check password?\n")
        sys.exit(1)

def _appli_func(argv):
    p = arg_parser_for_appli(argv[0])
    args = p.parse_args(argv[1:])
    if args.list:
        do_list_templates(args)
        sys.exit(0)

    if len(args.items) and args.items[0][0] not in string.ascii_uppercase:
        sys.stderr.write("Start command list with a command!\n")
        sys.exit(1)

    Applier(args).apply()

def waitfor_func():
    """Command-line entry point for appli-waitfor script.
    Parse args, invoke ops.
    """
    p = arg_parser_for_waitfor(sys.argv[0])
    args = p.parse_args(sys.argv[1:])
    Applier(args).wait()

def arg_parser_for_appli(prog):
    parser = arg_parser_common(prog, description="""Configure a Dell SMA
        appliance through its Aventail Management Console API.""")
    parser.add_argument('-t', '--timeout', dest='timeout',
                        default="30",
                        help="Wait this long for connection")
    parser.add_argument('-n', '--no-user-templates', dest='no_user_templates', action='store_true',
                        default=False,
                        help="Ignore user templates dir (--dir)")
    parser.add_argument('-l', '--list', dest='list', action='store_true',
                        default=False,
                        help="""List templates.
                        With no args, list standard templates and,
                        separately, user-defined templates and overrides.
                        With template names as arguments, list the contents of those templates.
                        With template names and name:value substitution parameters, list
                        those templates with substitutions completed as described. (To
                        view the results of default substitutions, provide a no-op substitution
                        parameter like "nonexistent:nothing")
                        """)
    return parser

def arg_parser_for_waitfor(prog):
    parser = arg_parser_common(prog, description="""Wait for a Dell SMA
        appliance to initialize after startup or address change.""")
    parser.add_argument('-t', '--timeout', dest='timeout',
                        default="120",
                        help="Wait this long for connection")
    return parser

def arg_parser_common(prog, description):
    epilog = """Config commands begin with an upper-case letter. They match
    API locations but with commas substituting for slashes. If the associated
    HTTP method is not PUT, it follows the command after two underscores, like
    'Version__GET'. Command strings are also template filenames. "Name:value"
    items list template substitutions. "Name=value" items list HTTP query
    parameters for the preceding command."""
    
    parser = argparse.ArgumentParser(prog=prog, description=description,
                                     epilog=epilog)
    parser.add_argument('-d', '--dir', dest="dir",
                        default=find_support_dir('appli'),
                        help="Path to support dir containing templates and settings")
    parser.add_argument('-hostname', '--hostname', dest="hostname",
                        help="Hostname (or IPv4 address) at which to find AMC")
    parser.add_argument('-u', '--url', dest="url",
                        default=None,
                        help="URL for AMC (-h overrides hostname)")
    parser.add_argument('-p', '--persist', dest="persist", action='store_true',
                        default=False,
                        help="Save URL details to settings in --dir (default ~/.appli)")
    parser.add_argument('-i', '--ignore-persist', dest="ignore_persist", action='store_true',
                        default=False,
                        help="Ignore ~/.appli/settings persist file")
    parser.add_argument('-P', '--prompt-password', dest='prompt_password', action='store_true',
                        default=False,
                        help="Prompt for password")
    parser.add_argument('-f', '--failok', dest='failok', action='store_true',
                        default=False,
                        help="Ignore APIFailure exceptions")
    parser.add_argument('-q', '--quiet', dest='quiet', action='store_true',
                        default=False,
                        help="Don't write diagnostic errors to stderr on failure")
    parser.add_argument('-v', '--verbose', dest='verbose', action='store_true',
                        default=False,
                        help="Write activity info to stderr")
    parser.add_argument('items', nargs="*",
                        help="Config commands and parameters")
    return parser

if __name__ == '__main__':
    appli_func()
